import os
import sys

sys.path.insert(0, os.path.abspath("."))
sys.path.insert(0, os.path.abspath(".."))
sys.path.insert(0, os.path.abspath(os.path.join("..", "..")))
sys.path.insert(0, os.path.abspath("../src"))
sys.path.insert(0, os.path.abspath("src"))
sys.path.insert(0, os.path.abspath("../src/nadzoring"))
sys.path.insert(0, os.path.abspath("src/nadzoring"))

project = "nadzoring"
author = "Alexeev Bronislav"
version = "0.1.0"
release = "0.1"
project_copyright = "2025, Alexeev Bronislaw"

autodoc_default_options = {
    "members": True,
    "undoc-members": True,
    "private-members": True,
    "special-members": "__init__",
}

extensions = [
    "sphinx.ext.autodoc",  # autodoc from docstrings
    "sphinx.ext.viewcode",  # links to source code
    "sphinx.ext.napoleon",  # support google and numpy docs style
    "sphinx.ext.todo",  # support TODO
    "sphinx.ext.coverage",  # check docs coverage
    "sphinx.ext.ifconfig",  # directives in docs
    "sphinx.ext.autosummary",  # generating summary for code
    "sphinx.ext.intersphinx",
    "sphinx.ext.githubpages",
]

pygments_style = "gruvbox-dark"

html_theme = "furo"  # theme
html_static_path = ["_static"]  # static dir
todo_include_todos = True  # include todo in docs
auto_doc_default_options = {"autosummary": True}

autodoc_mock_imports = []


def skip(app, what, name, obj, would_skip, options):
    if name == "__init__":
        return False
    return would_skip


def setup(app):
    app.connect("autodoc-skip-member", skip)



import logging
import sys
from logging import StreamHandler
from typing import ClassVar, Literal, TextIO

logger: logging.Logger = logging.getLogger("nadzoring")

logger.handlers.clear()
logger.propagate = False

DEBUG: Literal[10] = logging.DEBUG
INFO: Literal[20] = logging.INFO
WARNING: Literal[30] = logging.WARNING
ERROR: Literal[40] = logging.ERROR
CRITICAL: Literal[50] = logging.CRITICAL


class ColoredFormatter(logging.Formatter):
    """Custom formatter with colors"""

    grey: ClassVar[str] = "\x1b[38;20m"
    blue: ClassVar[str] = "\x1b[34;20m"
    green: ClassVar[str] = "\x1b[32;20m"
    yellow: ClassVar[str] = "\x1b[33;20m"
    red: ClassVar[str] = "\x1b[31;20m"
    bold_red: ClassVar[str] = "\x1b[31;1m"
    reset: ClassVar[str] = "\x1b[0m"

    format_str: ClassVar[str] = "%(asctime)s | %(levelname)-8s | %(name)s | %(message)s"
    simple_format_str: ClassVar[str] = "%(levelname)s: %(message)s"
    quiet_format_str: ClassVar[str] = "%(message)s"

    FORMATS: ClassVar[dict[int, str]] = {
        logging.DEBUG: blue + format_str + reset,
        logging.INFO: green + format_str + reset,
        logging.WARNING: yellow + format_str + reset,
        logging.ERROR: red + format_str + reset,
        logging.CRITICAL: bold_red + format_str + reset,
    }

    SIMPLE_FORMATS: ClassVar[dict[int, str]] = {
        logging.DEBUG: blue + simple_format_str + reset,
        logging.INFO: green + simple_format_str + reset,
        logging.WARNING: yellow + simple_format_str + reset,
        logging.ERROR: red + simple_format_str + reset,
        logging.CRITICAL: bold_red + simple_format_str + reset,
    }

    def __init__(
        self,
        fmt: str | None = None,
        datefmt: str | None = None,
        *,
        use_colors: bool = True,
        simple: bool = False,
    ) -> None:
        super().__init__(fmt, datefmt)
        self.use_colors = use_colors
        self.simple = simple

    def format(self, record: logging.LogRecord) -> str:
        if self.use_colors:
            if self.simple:
                formatter = logging.Formatter(
                    self.SIMPLE_FORMATS.get(record.levelno, self.simple_format_str),
                    self.datefmt,
                )
            else:
                formatter = logging.Formatter(
                    self.FORMATS.get(record.levelno, self.format_str), self.datefmt
                )
        elif self.simple:
            formatter = logging.Formatter(self.simple_format_str, self.datefmt)
        else:
            formatter = logging.Formatter(self.format_str, self.datefmt)

        return formatter.format(record)


def setup_cli_logging(
    *, verbose: bool = False, quiet: bool = False, no_color: bool = False
) -> None:
    """
    Setup logging for CLI mode.

    Args:
        verbose: verbose output (DEBUG)
        quiet: quiet mode (only ERROR and CRITICAL)
        no_color: disable colored output

    """
    logger.handlers.clear()

    console_handler: StreamHandler[TextIO] = logging.StreamHandler(sys.stderr)

    if verbose:
        logger.setLevel(logging.DEBUG)
        formatter = ColoredFormatter(
            datefmt="%Y-%m-%d %H:%M:%S", use_colors=not no_color, simple=False
        )
    elif quiet:
        logger.setLevel(logging.ERROR)
        formatter = ColoredFormatter(use_colors=not no_color, simple=True)
    else:
        logger.setLevel(logging.INFO)
        formatter = ColoredFormatter(use_colors=not no_color, simple=True)

    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)


def get_logger(name: str | None = None) -> logging.Logger:
    """
    Get child logger.

    Args:
        name: module name (for example, 'network_base.service_on_port')

    Returns:
        logging.Logger: logger child

    """
    if name:
        return logger.getChild(name)
    return logger



"""
Returns the default router(gateway) ip address on Linux and Windows machines.

On some Linux machines, it requires the net-tools package to be installed, as the
`route` command does not work without it (example: Ubuntu):
 $ sudo apt install net-tools
"""

from ipaddress import IPv4Address, IPv6Address
from logging import Logger
from platform import system
from socket import gethostbyname
from subprocess import check_output

from nadzoring.logger import get_logger

logger: Logger = get_logger(__name__)


def get_ip_from_host(hostname: str) -> str:
    """Attempts to retrieve an IP address from a domain name.

    If unsuccessful, returns the value passed to the function.

    Args:
        hostname (str): hostname address

    Returns:
        str: host
    """
    try:
        sock: str = gethostbyname(hostname)
    except Exception:
        return hostname
    else:
        return sock


def check_ipv4(hostname: str) -> str:
    """Checking whether the received value is an ip address.

    If the received value is not an ip address, an exception is thrown and the received
    value is passed to the function to obtain the address from the domain name. If the
    check is successful, the address is returned from the function.

    Args:
        ip (str): hostname address

    Returns:
        str: hostname
    """
    try:
        IPv4Address(hostname)
    except Exception:
        return get_ip_from_host(hostname)
    else:
        return hostname


def check_ipv6(hostname: str) -> str:
    """Checking whether the received value is an ip address.

    If the received value is not an ip address, an exception is thrown and the received
    value is passed to the function to obtain the address from the domain name. If the
    check is successful, the address is returned from the function.

    Args:
        ip (str): hostname address

    Returns:
        str: hostname
    """
    try:
        IPv6Address(hostname)
    except Exception:
        return get_ip_from_host(hostname)
    else:
        return hostname


def router_ip(*, ipv6: bool = False) -> str | None:
    """The OS version is determined, then, using the subprocess library,
    a command specific to each OS is executed to determine the router address.

    Then the received value is sent to check if it matches the address. If the received
    address is IPv4, it is returned from the function.
    If the address is a domain name, and there may be such cases,
    the received name is passed to the function where the attempt is performed.
    getting an ip address by domain name. For example, if the system
    uses pfsence.

    Args:
        ipv6 (bool, optional): use IPv6 instead of IPv4. Defaults to False.

    Returns:
        str | None: ip route result or None
    """

    if system() == "Linux":
        try:
            ip_route: str = str(
                check_output("route -n | grep UG", shell=True).decode().split()[1]  # noqa: S602, S607
            )
            return check_ipv4(ip_route) if not ipv6 else check_ipv6(ip_route)
        except Exception:
            logger.exception("Raised exception when router IP for Linux")
            return None
    elif system() == "Windows":
        try:
            ip_route: str = (
                check_output("route PRINT 0* -4 | findstr 0.0.0.0", shell=True)  # noqa: S602, S607
                .decode("cp866")
                .split()[-3]
            )
            return check_ipv4(ip_route) if not ipv6 else check_ipv6(ip_route)
        except Exception:
            logger.exception("Raised exception when router IP for Windows")
            return None
    else:
        logger.warning("System does not supported", extra=system())
    return None



"""
This script performs a standard ping of a specific address or domain name.
It works in the same way as the standard OS module. It does not ping addresses that
are protected from pinging, such as "codeby.net".
It can be used for a shallow check of addresses.
"""

from ipaddress import IPv4Address
from logging import Logger

from ping3 import ping

from nadzoring.logger import get_logger

logger: Logger = get_logger(__name__)


def ping_addr(addr: str) -> bool:
    """A normal ping of a specific ip address or domain.

    The original function returns if the domain or address is unavailable None.
    If successful, the time for which ping is being performed.
    This function returns Boolean values in the case of
    success or failure.

    Args:
        addr (str): address

    Returns:
        bool: is pinged
    """
    try:
        IPv4Address(addr)
    except Exception:
        if addr.startswith("http"):
            addr = addr.split("/")[2]
            if len(addr.split(".")) > 2:
                addr = ".".join(addr.split(".")[1:])

    try:
        return ping(addr) is not None
    except Exception:
        logger.exception("Raised exception when ping address")
        return False



import nox

# uv will handle any missing python versions
python_versions = ["3.12", "3.13", "3.14"]


@nox.session(python=python_versions, venv_backend="uv")
def test(session):
    """Run tests on specified Python versions."""
    # Install the package and test dependencies with uv
    session.run_always("uv", "pip", "install", ".", external=True)

    session.install(
        "pytest-xdist", "pytest-randomly", "pytest-sugar", "pytest-coverage"
    )

    # Run pytest with common options
    session.run(
        "pytest",
        "tests/",
        "--cov-fail-under=0",  # 100% coverage
        "-v",  # verbose output
        "-s",  # don't capture output
        "--tb=short",  # shorter traceback format
        "--strict-markers",  # treat unregistered markers as errors
        "-n",
        "auto",  # parallel testing
        *session.posargs,  # allows passing additional pytest args from command line
    )


@nox.session
def lint(session):
    session.install("ruff")
    session.run("ruff", "check", "src/nadzoring/")


@nox.session
def mutants(session):
    session.install("mutmut")
    session.run("mutmut", "run")



import csv
import functools
import io
import json
from collections.abc import Callable
from csv import DictWriter
from io import StringIO
from logging import Logger
from pathlib import Path
from time import time
from typing import Any, Literal, TextIO, cast

import click
from tabulate import tabulate

from nadzoring.logger import get_logger, setup_cli_logging
from nadzoring.network_base.network_params import network_param
from nadzoring.network_base.ping_address import ping_addr
from nadzoring.network_base.router_ip import (
    check_ipv4,
    check_ipv6,
    get_ip_from_host,
    router_ip,
)
from nadzoring.network_base.service_on_port import get_service_on_port

logger: Logger = get_logger("cli")


def colorize_result(result: dict[str, Any]) -> dict[str, Any]:
    """Add color coding to results"""
    colored: dict[str, str] = {}
    for key, value in result.items():
        if key == "IsPinged":
            if value == "yes":
                colored[key] = click.style(str(value), fg="green", bold=True)
            else:
                colored[key] = click.style(str(value), fg="red", bold=True)
        elif "Check" in key:
            if value == "passed":
                colored[key] = click.style(str(value), fg="green")
            elif value == "failed":
                colored[key] = click.style(str(value), fg="red")
            else:
                colored[key] = str(value)
        else:
            colored[key] = str(value)
    return colored


def print_results_table(results: dict[str, any]) -> None:
    if results:
        colored_results: list[dict[str, Any]] = [colorize_result(r) for r in results]
        print(
            tabulate(
                colored_results,
                showindex="never",
                headers="keys",
                tablefmt="simple_grid",
            )
        )
    else:
        print("No results.")


def print_csv_table(data: list[dict[str, Any]], file: TextIO | None = None) -> None:
    """Print data as CSV to console or file"""
    if not data:
        click.echo("No data to display")
        return None

    output: StringIO | TextIO = io.StringIO() if file is None else file
    writer: DictWriter[str] = csv.DictWriter(output, fieldnames=data[0].keys())

    writer.writeheader()
    writer.writerows(data)

    if file is None:
        click.echo(output.getvalue())
    return output.getvalue() if file is None else None


def save_results(data: Any, filename: str, fileformat: str) -> None:
    """Save results to file in specified format using pathlib"""
    try:
        file_path = Path(filename)

        file_path.parent.mkdir(parents=True, exist_ok=True)

        if fileformat == "json":
            with file_path.open("w", encoding="utf-8") as f:
                json.dump(data, f, indent=2, ensure_ascii=False)
            click.secho(f"✓ JSON results saved to {file_path}", fg="green")

        elif fileformat == "csv":
            with file_path.open("w", encoding="utf-8", newline="") as f:
                if data:
                    writer = csv.DictWriter(f, fieldnames=data[0].keys())
                    writer.writeheader()
                    writer.writerows(data)
            click.secho(f"✓ CSV results saved to {file_path}", fg="green")

        elif fileformat == "table":
            with file_path.open("w", encoding="utf-8") as f:
                if data:
                    f.write(tabulate(data, headers="keys", tablefmt="grid"))
            click.secho(f"✓ Table results saved to {file_path}", fg="green")

    except PermissionError:
        click.secho(
            f"✗ Permission denied: Cannot write to {filename}", fg="red", err=True
        )
    except OSError as e:
        click.secho(f"✗ OS error while saving results: {e}", fg="red", err=True)
    except Exception as e:
        click.secho(f"✗ Failed to save results: {e}", fg="red", err=True)


def common_logging_options[F: Callable[..., Any]](func: F) -> F:
    @click.option("--verbose", "-v", is_flag=True, help="Verbose output")
    @click.option("--quiet", "-q", is_flag=True, help="Quiet mode")
    @click.option("--no-color", is_flag=True, help="Disable colored output")
    @click.option(
        "--output",
        "-o",
        type=click.Choice(["table", "json", "csv"]),
        default="table",
        help="Output format",
    )
    @click.option("--save", type=click.Path(), help="Save results to file")
    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        verbose = kwargs.pop("verbose", False)
        quiet = kwargs.pop("quiet", False)
        no_color = kwargs.pop("no_color", False)
        output = kwargs.pop("output", False)
        save = kwargs.pop("save", False)
        setup_cli_logging(verbose=verbose, quiet=quiet, no_color=no_color)

        start_time = time()
        result = func(*args, **kwargs)
        elapsed = time() - start_time

        if output == "json":
            print(json.dumps(result))
        elif output == "table":
            print_results_table(result)
        elif output == "csv":
            print_csv_table(result)

        if save:
            save_results(result, save, output)

        if verbose:
            click.secho(
                f"\n⚡ Completed in {elapsed:.2f} seconds", fg="bright_black", dim=True
            )

        return result

    return cast(F, wrapper)


@click.group()
def cli() -> None:
    """FOSS tool for detecting website blocks, downdetecting and network analysis."""


@cli.group()
def network_base() -> None:
    """Network Base"""


@network_base.command()
@common_logging_options
@click.argument("addresses", type=str, nargs=-1, required=True)
def ping_address(addresses: tuple[str, ...]) -> list[dict[str, str]]:
    """Ping one or more addresses."""
    results: list[dict[str, str]] = []

    for address in addresses:
        is_pinged: Literal["no", "yes"] = "yes" if ping_addr(address) else "no"

        results.append({"Address": address, "IsPinged": is_pinged})

    return results


@network_base.command()
@common_logging_options
def get_network_params() -> list[dict[str, str | None] | None]:
    data: dict[str, str | None] | None = network_param()

    return [data]


@network_base.command()
@common_logging_options
@click.argument("hostnames", type=str, nargs=-1, required=True)
def get_ip_by_hostname(hostnames: tuple[str, ...]) -> None:
    """Get IPs for one or more hostname addresses."""
    results: list[dict[str, str | int]] = []

    router_ipv4: str | None = router_ip(ipv6=False)
    router_ipv6: str | None = router_ip(ipv6=True)

    for hostname in hostnames:
        ip: str = get_ip_from_host(hostname)
        ipv4_check: str = check_ipv4(hostname)
        ipv6_check: str = check_ipv6(hostname)

        if ip == hostname:
            logger.warning("Hostname %s is invalid", ip)

        results.append(
            {
                "Hostname": hostname,
                "IP Address": ip,
                "IPv4 Check": ipv4_check,
                "IPv6 Check": ipv6_check,
                "Router IPv4": router_ipv4 or "Not found",
                "Router IPv6": router_ipv6 or "Not found",
            }
        )

    return results


@network_base.command()
@common_logging_options
@click.argument("ports", type=int, nargs=-1, required=True)
def get_service_by_port(ports: tuple[int, ...]) -> None:
    """Get service names for one or more ports."""
    results: list[dict[str, str | int]] = []

    for port in ports:
        service: str = get_service_on_port(port)

        if service == "Unknown":
            logger.warning("Service is unknown for port %d; check port", port)

        results.append({"port": port, "service": service})

    return results


def main() -> None:
    """Entrypoint to CLI Application."""
    cli()
